from __future__ import annotations


from .enums import InteractionResponseType
from .slash.http import ModifiedSlashState
from .errors import InvalidEvent, WrongType
from .http import BetterRoute, get_message_payload, send_files
from .slash.errors import AlreadyDeferred, EphemeralDeletion
from .tools import EMPTY_CHECK, MISSING, All, deprecated, setup_logger, get
from .slash.types import ContextCommand, SlashCommand, SlashPermission, SlashSubcommand
from .components import ActionRow, Button, ComponentStore, LinkButton, SelectMenu, SelectOption, UseableComponent, make_component

import discord
from discord import utils
from discord.ext import commands
from discord.state import ConnectionState

from typing import Any, List, Union, Dict
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

logging = setup_logger("discord-ui")

__all__ = (
    'Message',
    'EphemeralMessage',
    'EphemeralResponseMessage',
    
    'ButtonInteraction',
    'PressedButton',    # deprecated

    'SelectInteraction',
    'SelectedMenu',     # deprecated

    'AutocompleteInteraction',
    'ChoiceGeneratorContext',   # deprecated

    'SlashInteraction',
    'SlashedCommand',           # deprecated

    'SubSlashInteraction',
    'SlashedSubCommand',        # deprecated

    'Interaction',
)

class InteractionType:
    PING                                =       Ping        =           1
    APPLICATION_COMMAND                 =      Command      =           2
    MESSAGE_COMPONENT                   =     Component     =           3
    APPLICATION_COMMAND_AUTOCOMPLETE    =    Autocomplete   =           4

class Interaction():
    def __init__(self, state, data, user=None, message=None) -> None:
        self._state: ModifiedSlashState = state

        self.deferred: bool = False
        self.responded: bool = False
        self._deferred_hidden: bool = False
        self._original_payload: dict = data

        self.author: Union[discord.Member, discord.User] = user
        """The user who created the interaction"""
        self.application_id: int = data["application_id"]
        """The ID of the bot application"""
        self.token: str = data["token"]
        """The token for responding to the interaction"""
        self.id: int = int(data["id"])
        """The id of the interaction"""
        self.type: int = data["type"]
        """The type of the interaction. See :class:`~InteractionType` for more information"""
        self.version: int = data["version"]
        self.data: dict = data["data"]
        """The passed data of the interaction"""
        self.channel_id: int = int(data.get("channel_id")) if data.get("channel_id") is not None else None
        """The channel-id where the interaction was created"""
        self.guild_id: int = int(data["guild_id"]) if data.get("guild_id") is not None else None
        """The guild-id where the interaction was created"""
        self.message: Message = message
        """The message in which the interaction was created"""

    @property
    def created_at(self):
        """The interaction's creation time in UTC"""
        return utils.snowflake_time(self.id)
    
    @property
    def guild(self) -> discord.Guild:
        """The guild where the interaction was created"""
        return self._state._get_guild(self.guild_id)
    @property
    def channel(self) -> Union[discord.abc.GuildChannel, discord.abc.PrivateChannel]:
        """The channel where the interaction was created"""
        return self._state.get_channel(self.channel_id) or self._state.get_channel(self.author.id)

    async def defer(self, hidden=False):
        """
        This will acknowledge the interaction. This will show the (*Bot* is thinking...) Dialog

        .. note::
            
            This function should be used if the bot needs more than 15 seconds to respond
        
        Parameters
        ----------
        hidden: :class:`bool`, optional
            Whether the loading thing should be only visible to the user; default False.
        
        """
        if self.deferred:
            logging.error(AlreadyDeferred())
            return

        payload = None
        if hidden is True:
            payload = {"flags": 64}
            self._deferred_hidden = True
        
        await self._state.slash_http.respond_to(self.id, self.token, InteractionResponseType.Deferred_channel_message, payload)
        self.deferred = True

    async def respond(self, content=None, *, tts=False, embed=None, embeds=None, file=None, files=None, nonce=None,
    allowed_mentions=None, mention_author=None, components=None, delete_after=None, listener=None, 
    hidden=False, ninja_mode=False) -> Union['Message', 'EphemeralMessage']:
        """
        Responds to the interaction
        
        Parameters
        ----------
        content: :class:`str`, optional
            The raw message content
        tts: :class:`bool`
            Whether the message should be send with text-to-speech
        embed: :class:`discord.Embed`
            Embed rich content
        embeds: List[:class:`discord.Embed`]
            A list of embeds for the message
        file: :class:`discord.File`
            The file which will be attached to the message
        files: List[:class:`discord.File`]
            A list of files which will be attached to the message
        nonce: :class:`int`
            The nonce to use for sending this message
        allowed_mentions: :class:`discord.AllowedMentions`
            Controls the mentions being processed in this message
        mention_author: :class:`bool`
            Whether the author should be mentioned
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of message components to be included
        delete_after: :class:`float`
            After how many seconds the message should be deleted, only works for non-hiddend messages; default MISSING
        listener: :class:`Listener`
            A component-listener for this message
        hidden: :class:`bool`
            Whether the response should be visible only to the user 
        ninja_mode: :class:`bool`
            If true, the client will respond to the button interaction with almost nothing and returns nothing
        
        Returns
        -------
        :class:`~Message` | :class:`~EphemeralMessage`
            Returns the sent message
        """
        if ninja_mode is True or all(y in [None, False] for x, y in locals().items() if x not in ["self"]):
            try:
                await self._state.slash_http.respond_to(self.id, self.token, InteractionResponseType.Deferred_message_update)
                self.responded = True
                return
            except discord.errors.HTTPException as x:
                if "value must be one of (4, 5)" in str(x).lower():
                    logging.error(str(x) + "\n" + "The 'ninja_mode' parameter is not supported for slash commands!")
                    ninja_mode = False
                else:
                    raise x

        if self.responded is True:
            return await self.send(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, 
                components=components, listener=listener, hidden=hidden
            )

        if components is None and listener is not None:
            components = listener.to_components()        
        payload = get_message_payload(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, 
            mention_author=mention_author, components=components
        )
        
        if self._deferred_hidden is hidden:
            if self._deferred_hidden is False and hidden is True:
                logging.warning("Your response should be hidden, but the interaction was deferred public. This results in a public response.")
            if self._deferred_hidden is True and hidden is False:
                logging.warning("Your response should be public, but the interaction was deferred hidden. This results in a hidden response.")
        hide_message = self._deferred_hidden or not self.deferred and hidden is True


        r = None
        if delete_after is not None and hide_message is True:
            raise EphemeralDeletion()

        if hide_message:
            payload["flags"] = 64
        
        if self.deferred:
            route = BetterRoute("PATCH", f'/webhooks/{self.application_id}/{self.token}/messages/@original')
            if file is not None or files is not None:
                await send_files(route=route, files=files or ([file] if file is not None else None), payload=payload, http=self._state.http)
            else:
                await self._state.http.request(route, json=payload)    
        else:
            await self._state.slash_http.respond_to(self.id, self.token, InteractionResponseType.Channel_message, payload, files=files or [file] if file is not None else None)
        self.responded = True
        
        r = await self._state.http.request(BetterRoute("GET", f"/webhooks/{self.application_id}/{self.token}/messages/@original"))
        if hide_message is True:
            msg = EphemeralMessage(state=self._state, channel=self.channel, data=r, application_id=self.application_id, token=self.token)
        else:
            msg = await getMessage(self._state, data=r, response=False)
        if listener is not None:
            listener._start(msg)
        if delete_after is not None:
            await msg.delete(delete_after)
        return msg
    async def send(self, content=None, *, tts=None, embed=None, embeds=None, file=None, files=None, nonce=None,
        allowed_mentions=None, mention_author=None, components=None, delete_after=None, listener=None, hidden=False,
        force=False
    ) -> Union[Message, EphemeralMessage]:
        """
        Sends a message to the interaction using a webhook
        
        Parameters
        ----------
        content: :class:`str`, optional
            The raw message content
        tts: :class:`bool`, optional
            Whether the message should be send with text-to-speech
        embed: :class:`discord.Embed`, optional
            Embed rich content
        embeds: List[:class:`discord.Embed`], optional
            A list of embeds for the message
        file: :class:`discord.File`, optional
            The file which will be attached to the message
        files: List[:class:`discord.File`], optional
            A list of files which will be attached to the message
        nonce: :class:`int`, optional
            The nonce to use for sending this message
        allowed_mentions: :class:`discord.AllowedMentions`, optional
            Controls the mentions being processed in this message
        mention_author: :class:`bool`, optional
            Whether the author should be mentioned
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of message components to be included
        delete_after: :class:`float`, optional
            After how many seconds the message should be deleted, only works for non-hiddend messages; default MISSING
        listener: :class:`Listener`, optional
            A component-listener for this message
        hidden: :class:`bool`, optional
            Whether the response should be visible only to the user 
        ninja_mode: :class:`bool`, optional
            If true, the client will respond to the button interaction with almost nothing and returns nothing
        force: :class:`bool`, optional
            Whether sending the follow-up message should be forced.
            If ``False``, then a follow-up message will only be send if ``.responded`` is True; default False
        
        Returns
        -------
        :class:`~Message` | :class:`EphemeralMessage`
            Returns the sent message
        """
        if force is False and self.responded is False:
            return await self.respond(content=content, tts=tts, embed=embed, embeds=embeds, file=file, files=files, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components, delete_after=delete_after, listener=listener, hidden=hidden)

        if components is None and listener is not None:
            components = listener.to_components()
        payload = get_message_payload(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, mention_author=mention_author, components=components)
        
        if hidden:
            payload["flags"] = 64

        route = BetterRoute("POST", f'/webhooks/{self.application_id}/{self.token}')
        if file is not None or files is not None:
            r = await send_files(route=route, files=files or ([file] if file is None else None), payload=payload, http=self._state.http)
        else:
            r = await self._state.http.request(route, json=payload)

        if hidden is True:
            msg = EphemeralMessage(state=self._state, channel=self._state.get_channel(int(r["channel_id"])), data=r, application_id=self.application_id, token=self.token)
        else:
            msg = await getMessage(self._state, r, response=False)
        if delete_after is not None:
            await msg.delete(delete_after)
        if listener is not None:
            listener._start(msg)
        return msg
    def _handle_auto_defer(self, auto_defer):
        self.deferred = auto_defer[0]
        self._deferred_hidden = auto_defer[1]

class AutocompleteInteraction(Interaction):
    """Autocomplete interaction"""
    def __init__(self, command, state, data, options, user=None) -> None:
        Interaction.__init__(self, state, data, user=user)
        self.focused_option: dict = options[get(options, check=lambda x: options[x].get("focused", False))]
        """The option for which the choices should be generated"""
        self.value_query: Union[str, int] = self.focused_option["value"]
        """The current input of the focused option"""
        self.selected_options: Dict[str, Any] = {options[x]["name"]: options[x]["value"] for x in options}
        """All the options that were already selected. Format: ``{"option name": value}``"""
        self.command: Union[SlashInteraction, SlashInteraction, ContextInteraction] = command
        """The slash command for which the choices should be generated"""

    async def defer(self, *args, **kwargs):
        """Cannot defer this type of interaction"""
        raise NotImplementedError()
    async def respond(self, *args, **kwargs):
        """Response will be made automatically with the choices that are returned"""
        raise NotImplementedError()
    async def send(self, *args, **kwargs):
        """Cannot send followup message to this type of interaction"""
        raise NotImplementedError()
class ChoiceGeneratorContext(AutocompleteInteraction):
    """Deprecated, please use :class:`AutocompleteInteraction` instead"""
    ...

class ComponentInteraction(Interaction):
    """A received component"""
    def __init__(self, state, data, user, message) -> None:
        Interaction.__init__(self, state, data, user=user, message=message)
        self.component: UseableComponent = UseableComponent(data["data"]["component_type"])
        self.component._custom_id = data["data"]["custom_id"]
class ComponentContext(ComponentInteraction):
    """Deprecated, please use :class:`ComponentInteraction` instead"""
    ...

class SelectInteraction(Interaction):
    """An interaction that was created by a :class:`~SelectMenu`"""
    def __init__(self, data, user, s, msg, client) -> None:
        Interaction.__init__(self, client._connection, data, user, msg)
        
        self.component: SelectMenu = s
        self.bot: commands.Bot = client
        self.selected_options: List[SelectOption] = []
        """The list of the selected options"""
        self.selected_values: List[str] = []
        """The list of raw values which were selected"""
        for val in data["data"]["values"]:
            for x in self.component.options:
                if x.value == val:
                    self.selected_options.append(x)
                    self.selected_values.append(x.value)
        self.author: discord.Member = user
        """The user who selected the value"""
class SelectedMenu(SelectInteraction):
    """deprecated, please use :class:`SelectInteraction` instead"""
    ...

class ButtonInteraction(Interaction):
    """An interaction that was created by a :class:`~Button`"""
    def __init__(self, data, user, b, message, client) -> None:
        Interaction.__init__(self, client._connection, data, user, message)
        self.component: Button = b
        """The component that created the interaction"""
        self.bot: commands.Bot = client
        self.author: discord.Member = user
        """The user who pressed the button"""
class PressedButton(ButtonInteraction):
    """deprecated, please use :class:`ButtonInteraction` instead"""
    ...

class SlashInteraction(Interaction):
    """An interaction created by a :class:`~SlashCommand`"""
    def __init__(self, client, command: SlashCommand, data, user, args = None) -> None:
        Interaction.__init__(self, client._connection, data, user)
        self.command: SlashCommand = command
        """The original command instance that was used. If you change things here, the changes will be applied globally"""
        self.bot: commands.Bot = client
        self.author: discord.Member = user
        """The user who used the command"""
        self.args: Dict[str, Union[str, int, bool, discord.Member, discord.TextChannel, discord.Role, float]] = args
        """The options that were received"""
        self.permissions: SlashPermission = command.guild_permissions.get(self.guild_id) if command.guild_permissions is not None else None
        """The permissions for this guild"""
class SlashedCommand(SlashInteraction):
    """deprecated, please use :class:`SlashInteraction` instead"""
    ...

class SubSlashInteraction(SlashInteraction):
    """An interaction created by a :class:`~SlashSubcommand`"""

    command: SlashSubcommand
    def __init__(self, client, command, data, user, args = None) -> None:
        SlashInteraction.__init__(self, client, command, data, user, args)
class SlashedSubCommand(SubSlashInteraction):
    """deprecated, please use ``SubSlashInteraction`` instead"""
    ...

class ContextInteraction(Interaction):
    """An interaction created by a :class:`~MessageCommand` or a :class:`~UserCommand`"""
    def __init__(self, client, command: ContextCommand, data, user, target) -> None:
        Interaction.__init__(self, client._connection, data, user)
        self.command: ContextCommand = command
        """The original command instance that was used"""
        self.bot: commands.Bot = client
        self.target: Union[Message, Union[discord.Member, discord.User]] = target
        """The target object on which the interaction was used"""
        self.permissions: SlashPermission = command.guild_permissions.get(self.guild_id) if command.guild_permissions is not None else None 
        """The permissions for this guild"""
        

async def getMessage(state: discord.state.ConnectionState, data, response=True):
    """
    Async function to get the response message

    Parameters
    -----------------
    state: :class:`discord.state.ConnectionState`
        The discord bot client
    data: :class:`dict`
        The raw data
    user: :class:`discord.User`
        The User which pressed the button
    response: :class:`bool`
        Whether the Message returned should be of type `ResponseMessage` or `Message`
    channel: :class:`discord.Messageable`
        An alternative channel that will be used when no channel was found

    Returns
    -------
    :class:`~Message` | :class:`~EphemeralMessage`
        The sent message
    """
    msg_base = data.get("message", data)

    channel = state.get_channel(int(data["channel_id"])) or state.get_channel(int(msg_base["author"]["id"]))
    if response:
        if msg_base["flags"] == 64:
            return EphemeralResponseMessage(state=state, channel=channel, data=data.get("message", data))
        return Message(state=state, channel=channel, data=msg_base)

    if msg_base["flags"] == 64:
        return EphemeralMessage(state=state, channel=channel, data=msg_base)
    return Message(state=state, channel=channel, data=msg_base)

class Message(discord.Message):
    """A :class:`discord.Message` optimized for components"""

    _state: ConnectionState
    def __init__(self, *, state, channel, data):
        self.__slots__ = discord.Message.__slots__ + ("components",)
        discord.Message.__init__(self, state=state, channel=channel, data=data)
        self.components = ComponentStore()
        """The components in the message"""

        self._update_components(data)

    # region attributes
    @property
    @deprecated(".components.buttons")
    def buttons(self) -> List[Union[Button, LinkButton]]:
        """The button components in the message"""
        return self.components.buttons
    @property
    @deprecated(".components.selects")
    def select_menus(self) -> List[SelectMenu]:
        """The select menus components in the message"""
        return self.components.selects
    @property
    @deprecated(".components.get_rows()")
    def action_rows(self) -> List[ActionRow]:
        return self.components.get_rows()

    def _update_components(self, data):
        """Updates the message components"""
        if data.get("components") is None:
            self.components = ComponentStore()
            return
        self.components = ComponentStore()
        if len(data["components"]) == 0:
            pass
        elif len(data["components"]) > 1:
            # multiple lines
            for componentWrapper in data["components"]:
                # newline
                for index, com in enumerate(componentWrapper["components"]):
                    self.components.append(make_component(com, index==0))
        elif len(data["components"][0]["components"]) > 1:
            # All inline
            for index, com in enumerate(data["components"][0]["components"]):
                self.components.append(make_component(com, index==0))
        else:
            # One button
            component = data["components"][0]["components"][0]
            self.components.append(make_component(component))
    def _update(self, data):
        super()._update(data)
        self._update_components(data)

    async def edit(self, content=MISSING, *, embed=MISSING, embeds=MISSING, attachments=MISSING, suppress=MISSING, 
        delete_after=MISSING, allowed_mentions=MISSING, components=MISSING):
        """Edits the message and updates its properties

        .. note::

            If a paremeter is `None`, the attribute will be removed from the message

        Parameters
        ----------------
        content: :class:`str`
            The new message content
        embed: :class:`discord.Embed`
            The new embed of the message
        embeds: List[:class:`discord.Embed`]
            The new list of discord embeds
        attachments: List[:class:`discord.Attachment`]
            A list of new attachments
        supress: :class:`bool`
            Whether the embeds should be shown
        delete_after: :class:`float`
            After how many seconds the message should be deleted
        allowed_mentions: :class:`discord.AllowedMentions`
            The mentions proceeded in the message
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
            A list of components to be included the message
        """
        payload = get_message_payload(content, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, attachments=attachments, suppress=suppress, flags=self.flags.value, components=components)
        data = await self._state.http.edit_message(self.channel.id, self.id, **payload)
        self._update(data)

        if delete_after is not MISSING:
            await self.delete(delay=delete_after)

    async def disable_components(self, index=All, disable=True, **fields):
        """Disables component(s) in the message
        
        Parameters
        ----------
        index: :class:`int` | :class:`str` | :class:`range` | List[:class:`int` | :class:`str`], optional
            Index(es) or custom_id(s) for the components that should be disabled or enabled; default all components
        disable: :class:`bool`, optional
            Whether to disable (``True``) or enable (``False``) components; default True
        ``**fields``
            Other parameters for editing the message (like `content=`, `embed=`)
        """
        self.components.disable(index, disable)
        await self.edit(components=self.components, **fields)

    async def wait_for(self, event_name: Literal["select", "button", "component"], client, custom_id=None, by=None, check=EMPTY_CHECK, timeout=None) -> Union[ButtonInteraction, SelectInteraction, ComponentContext]:
        """Waits for a message component to be invoked in this message

        Parameters
        -----------
        event_name: :class:`str`
            The name of the event which will be awaited [``"select"`` | ``"button"`` | ``"component"``]
            
            .. note::

                ``event_name`` must be ``select`` for a select menu selection, ``button`` for a button press and ``component`` for any component

        client: :class:`discord.ext.commands.Bot`
            The discord client
        custom_id: :class:`str`, Optional
            Filters the waiting for a custom_id
        by: :class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`, Optional
            The user or the user id by that has to create the component interaction
        check: :class:`function`, Optional
            A check that has to return True in order to break from the event and return the received component
                The function takes the received component as the parameter
        timeout: :class:`float`, Optional
            After how many seconds the waiting should be canceled. 
            Throws an :class:`asyncio.TimeoutError` Exception

        Raises
        ------
        :class:`discord_ui.errors.InvalidEvent`
            The event name passed was invalid 

        Returns
        --------
        :class:`~ButtonInteraction` | :class:`~SelectInteraction`
            The component that was waited for

        Example
        -------

        .. code-block::

            # send a message with comoponents
            msg = await ctx.send("okay", components=[Button(custom_id="a_custom_id", ...)])
            try:
                # wait for the button
                btn = await msg.wait_for("button", client, "a_custom_id", by=ctx.author, timeout=20)
                # send response
                btn.respond()
            except asyncio.TimeoutError:
                # no button press was received in 20 seconds timespan
        """
        def _check(com):
            if com.message.id == self.id:
                statements = []
                if custom_id is not None:
                    statements.append(com.custom_id == custom_id)
                if by is not None:
                    statements.append(com.author.id == (by.id if hasattr(by, "id") else int(by)))
                if check is not None:
                    statements.append(check(com))
                return all(statements)
            return False
        if not isinstance(client, commands.Bot):
            raise WrongType("client", client, "discord.ext.commands.Bot")
        
        if event_name.lower() == "button":
            return await client.wait_for('button', check=_check, timeout=timeout)
        if event_name.lower() == "select":
            return await client.wait_for("select", check=_check, timeout=timeout)
        if event_name.lower() == "component":
            return await client.wait_for("component", check=_check, timeout=timeout)
        
        raise InvalidEvent(event_name, ["button", "select", "component"])

    async def put_listener(self, listener):
        """Adds a listener to this message and edits the message if the components of the listener are missing in this message
        
        Parameters
        ----------
        listener: :class:`Listener`
            The listener which should be put to the message
        
        """
        if len(self.components) == 0:
            await self.edit(components=listener.to_components())
        self.attach_listener(listener)       
    def attach_listener(self, listener):
        """Attaches a listener to this message after it was sent
        
        Parameters
        ----------
        listener: :class:`Listener`
            The listener that should be attached
        
        """
        listener._start(self)
    def remove_listener(self):
        """Removes the listener from this message"""
        try:
            del self._state._component_listeners[str(self.id)]
        except KeyError:
            pass

class EphemeralMessage(Message):
    """Represents a hidden (ephemeral) message"""
    def __init__(self, state, channel, data, application_id=None, token=None):
        #region fix reference keyerror
        if data.get("message_reference"):
            if data["message_reference"].get("channel_id") is None:
                data["message_reference"]["channel_id"] = str(channel.id)
        #endregion
        Message.__init__(self, state=state, channel=channel, data=data)
        self._application_id = application_id
        self._interaction_token = token
    async def edit(self, *args, **fields):
        r = BetterRoute("PATCH", f"/webhooks/{self._application_id}/{self._interaction_token}/messages/{self.id}")
        self._update(await self._state.http.request(r, json=get_message_payload(*args, **fields)))        
    async def delete(self):
        """Override for delete function that will throw an exception"""
        raise EphemeralDeletion()

class EphemeralResponseMessage(Message):
    """A ephemeral message which was created from an interaction
    
    .. important::

        Methods like `.edit()`, which change the original message, need a `token` paremeter passed in order to work
    """
    def __init__(self, *, state, channel, data):
        Message.__init__(self, state=state, channel=channel, data=data)

    async def edit(self, token, *args, **fields):
        """Edits the message
        
        Parameters
        ----------
        token: :class:`str`
            The token of the interaction with wich this ephemeral message was sent
        fields: :class:`kwargs`
            The fields to edit (ex. `content="...", embed=..., attachments=[...]`)

        Example

        .. code-block::

            async def testing(ctx):
                msg = await ctx.send("hello hidden world", components=[Button("test")])
                btn = await msg.wait_for("button", client)
                await btn.message.edit(ctx.token, content="edited", components=None)
    
        """
        route = BetterRoute("PATCH", f"/webhooks/{self.interaction.application_id}/{token}/messages/{self.id}")
        self._update(await self._state.http.request(route, json=get_message_payload(*args, **fields)))
    async def delete(self):
        """Override for delete function that will throw an exception"""
        raise EphemeralDeletion()
    async def disable_components(self, token, disable = True, **fields):
        """Disables all component in the message
        
        Parameters
        ----------
        disable: :class:`bool`, optional
            Whether to disable (``True``) or enable (``False``) all components; default True
        
        """
        self.components.disable(disable=disable)
        await self.edit(token, components=self.components, **fields)


"""
discord_ui.slash.ext
~~~~~~~~~~~~~~~~~~~~~~
An extension module to the libary that has some useful decorators and functions 
for application-commands.

.. code-block::

    from discord_ui import ext

"""

from .builder import *
from .command_decorators import *


from .cogs import BaseCallable, InteractionableCog, ListeningComponent
from .http import get_message_payload, BetterRoute, send_files
from .tools import MISSING, EMPTY_CHECK, _none, _or, deprecated, setup_logger, get
from .errors import MissingListenedComponentParameters, WrongType
from .components import Button, Component, SelectMenu

from .slash.http import SlashHTTP
from .slash.errors import NoAsyncCallback
from .slash.tools import AdditionalType, ParseMethod, handle_options, handle_thing
from .slash.types import (
    CommandCache, OptionType,
    MessageCommand , SlashCommand, SlashSubcommand, UserCommand
)

from .receive import (
    AutocompleteInteraction, ComponentContext, 
    Interaction, InteractionType,
    ButtonInteraction, SelectInteraction,
    SlashInteraction, SubSlashInteraction, ContextInteraction,
    getMessage, Message
)
from .listener import Listener
from .override import override_dpy as override_it
from .enums import CommandType, InteractionResponseType, ComponentType

from .override import override_dpy as override_it
from .listener import Listener
from .enums import InteractionResponseType, ComponentType


import discord
from discord.ext import commands

import json
import inspect
import asyncio
import contextlib
from typing import Any, Callable, Coroutine, Dict, List, Tuple, Union, TypeVar
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

logging = setup_logger(__name__)

__all__ = (
    'UI',
    'Slash',
    'Components',
)

# Command Type
_C = TypeVar("_C")

class Slash():
    """
    A class for using slash commands
    
    Parameters
    ----------
    client: :class:`commands.Bot`
        The bot client

    parse_method: :class:`bool`, optional
        How received option data should be treated; Default ``ParseMethod.AUTO``

    delete_unused: :class:`bool`, optional
        Whether the commands that are not registered by this slash extension should be deleted in the api; Default ``False``

    sync_on_cog: :class:`bool`, optional
        Whether the slashcommands should be updated whenever a new cog is added or removed; Default ``False``

    wait_sync: :class:`float`, optional
        How many seconds will be waited until the commands are going to be synchronized; Default ``1``

    auto_defer: Tuple[:class:`bool`, :class:`bool`]
        Settings for the auto-defer; Default ``(True, False)``

        ``[0]``: Whether interactions should be deferred automatically

        ``[1]``: Whether the deferration should be hidden (True) or public (False)

    Example
    ------------------
    Example for using slash commands

    .. code-block::

        ...
        # Your bot declaration and everything
        slash = Slash(client)

    For creating a slash command use

    .. code-block::

        ...
        @slash.command(name="my_command", description="this is my slash command", options=[SlashOption(str, "option", "this is an option")])
        async def command(ctx: SlashInteraction):
            ...
    
    And for subcommand groups use

    .. code-block::

        ...
        @slash.subcommand_group(base_names=["base", "group"], name="sub", description="this is a sub command group")
        async def subgroup(ctx: SubSlashInteraction):
            ...
        

    """
    def __init__(self, client, parse_method = ParseMethod.AUTO, auto_sync=True, delete_unused = False, sync_on_cog=False, wait_sync = 1, auto_defer = False) -> None:
        """
        Creates a new slash command thing
        
        Example
        ```py
        Slash(client)
        ```
        """
        self.ready = False
        self.parse_method: int = parse_method
        self.delete_unused: bool = delete_unused
        self.sync_on_cog: bool = sync_on_cog
        self.wait_sync: float = wait_sync
        self.auto_defer: Tuple[bool, bool] = (auto_defer, False) if isinstance(auto_defer, bool) else auto_defer
        self.auto_sync = auto_sync

        self._discord: commands.Bot = client
        self.http: SlashHTTP = None # set it when bot is connected
        self._discord._connection.slash_http = None # set when bot is connected
        self.commands = CommandCache(self._discord)
        
        if discord.__version__.startswith("2"):
            self._discord.add_listener(self._on_slash_response, "on_socket_raw_receive")
        elif discord.__version__.startswith("1"):
            self._discord.add_listener(self._on_slash_response, 'on_socket_response')

        old_add = self._discord.add_cog
        def add_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            if isinstance(cog, str):
                cog = self._discord.get_cog(cog)
            if not isinstance(cog, InteractionableCog):
                # adding attributees to cog form InteractionableCog
                for s in InteractionableCog.__custom_slots__:
                    setattr(cog, "s", getattr(InteractionableCog, s))
            for com in self._get_cog_commands(cog):
                com.cog = cog
                self.commands.add(com)
            old_add(*args, **kwargs)
            if self.ready and self.sync_on_cog is True:
                self._discord.loop.create_task(self.commands.sync(self.delete_unused))
        self._discord.add_cog = add_cog_override

        old_remove = self._discord.remove_cog
        def remove_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            if isinstance(cog, str):
                cog = self._discord.get_cog(cog)
            for com in self._get_cog_commands(cog):
                com.cog = cog
                self.commands.remove(com)
            old_remove(*args, **kwargs)
            if self.ready and self.sync_on_cog is True:
                self._discord.loop.create_task(self.commands.sync(self.delete_unused))
        self._discord.remove_cog = remove_cog_override
        
        async def on_connect():
            self.http = SlashHTTP(self._discord)
            self._discord._connection.slash_http = self.http
            self.ready = True
            if self.auto_sync is False:
                return
            await asyncio.sleep(_or(self.wait_sync, 1))
            await self.commands.sync(self.delete_unused)
            # await self.sync_commands(self.delete_unused)
        self._discord.add_listener(on_connect)

    @deprecated("commands.sync")
    async def sync_commands(self, delete_unused=False):
        """deprecated, use ``commands.sync`` instead"""
        return await self.commands.sync(delete_unused)
    async def _on_slash_response(self, msg):
        if discord.__version__.startswith("2"):
            if isinstance(msg, bytes):
                raise NotImplementedError("decompressing was removed! Please upgrade your discord.py version")
            if isinstance(msg, str):
                msg = json.loads(msg)
        if msg["t"] != "INTERACTION_CREATE":
            return
        data = msg["d"]

        # filter out any interaction that is not a application command interaction
        if int(data["type"]) not in [InteractionType.PING, InteractionType.APPLICATION_COMMAND, InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE]:
            return

        # get the author
        user = discord.Member(data=data["member"], guild=self._discord._connection._get_guild(int(data["guild_id"])), state=self._discord._connection) if data.get("member") is not None else discord.User(state=self._discord._connection, data=data["user"])
        # as stated in https://github.com/discord-py-ui/discord-ui/issues/94, .author.send won't work because if no dm_channel was opened for the user 
        if user.dm_channel is None:
            await user.create_dm()

        # things for autocomplete
        if int(data["type"]) == InteractionType.APPLICATION_COMMAND_AUTOCOMPLETE:
            raw_options = {}
            """The options that were already selected"""
            command = None
            """The original command"""
            # if command is not slash command (this cenario is not possible, but you never know)
            if CommandType(data["data"]["type"]) is not CommandType.Slash:
                return
            command = self.commands.get_command_for(data)
            # if the command is not a subcommand
            if not (data["data"].get("options") and data["data"]["options"][0]["type"] in [OptionType.SUB_COMMAND, OptionType.SUB_COMMAND_GROUP]):
                raw_options = []
                if data["data"].get("options") is not None:
                    options = await handle_options(data, data["data"]["options"], self.parse_method, self._discord)
                    raw_options = data["data"]["options"]
            # if command is a subcommand
            else:
                if command:
                    op = data["data"]["options"][0]
                    while op["type"] != 1:
                        op = op["options"][0]
                    fixed_options = op.get("options", [])
                    raw_options = fixed_options
                    options = await handle_options(data, fixed_options, self.parse_method, self._discord)

            if command is None:
                logging.warning("no slashcommand handler found for " + data["data"]["name"])
                return
            
            parsed_options = {
                x["name"]: {
                    "name": x["name"], 
                    "value": await handle_thing(
                            x["value"], x["type"], data, self.parse_method, self._discord
                    )
                } | (
                    {"focused": True} if x.get("focused") else {}
                ) 
                for x in raw_options
            }
            choice_ctx = AutocompleteInteraction(command, self._discord._connection, data, parsed_options, user) 
            return await self.http.respond_to(choice_ctx.id, choice_ctx.token, InteractionResponseType.Autocomplete_result, {
                "choices": [
                    (
                        {"name": x[0], "value": x[1]} if isinstance(x, tuple) else x
                    ) for x in (
                        await (
                            get(command.options, choice_ctx.focused_option["name"], lambda x: getattr(x, "name", None)
                            )
                        ).choice_generator( *(
                            [command.cog, choice_ctx] 
                                if hasattr(command, "cog") and command.cog is not None else 
                            [choice_ctx]
                        ) )
                    )
                ]
            })

        interaction = Interaction(self._discord._connection, data, user)
        if self.auto_defer[0] is True:
            await interaction.defer(self.auto_defer[1])
        self._discord.dispatch("interaction_received", interaction)


        command = self.commands.get_command_for(data)
        
        # region basic-commands
        # slash command
        if CommandType(data["data"]["type"]) is CommandType.Slash and not (data["data"].get("options") and data["data"]["options"][0]["type"] in [OptionType.SUB_COMMAND, OptionType.SUB_COMMAND_GROUP]):
            if command is not None:
                options = {}
                if data["data"].get("options") is not None:
                    options = await handle_options(data, data["data"]["options"], self.parse_method, self._discord)
                context = SlashInteraction(self._discord, command=command, data=data, user=user, args=options)
                # Handle autodefer
                context._handle_auto_defer(self.auto_defer)
                self._discord.dispatch("slash_command", context)
                if hasattr(command, "invoke"):
                    await command.invoke(context, **options)
                elif command.callback is not None:
                    await command.callback(context, **options)
                return
        # UserContext command
        elif CommandType(data["data"]["type"]) is CommandType.User:
            if command is not None:
                member = await handle_thing(data["data"]["target_id"], OptionType.MEMBER, data, self.parse_method, self._discord)
                context = ContextInteraction(self._discord, command=command, data=data, user=user, target=member)
                # Handle autodefer
                context._handle_auto_defer(self.auto_defer)

                self._discord.dispatch("context_command", context, member)
                if command.callback is not None:
                    if hasattr(command, "invoke"):
                        await command.invoke(context, member)
                    else:
                        await command.callback(context, member)
                return
        # MessageContext command
        elif CommandType(data["data"]["type"]) is CommandType.Message:
            if command is not None:
                message = await handle_thing(data["data"]["target_id"], AdditionalType.MESSAGE, data, self.parse_method, self._discord)
                context = ContextInteraction(self._discord, command=command, data=data, user=user, target=message)
                # Handle autodefer
                context._handle_auto_defer(self.auto_defer)
                
                self._discord.dispatch("context_command", context, message)
                if command.callback is not None:
                    if hasattr(command, "invoke"):
                        await command.invoke(context, message)
                    else:
                        await command.callback(context, message)
                return
        #endregion

        # subcommands
        fixed_options = []
        if command:
            op = data["data"]["options"][0]
            while op["type"] != 1:
                op = op["options"][0]
            fixed_options = op.get("options", [])
            options = await handle_options(data, fixed_options, self.parse_method, self._discord)

            context = SubSlashInteraction(self._discord, command, data, user, options)
            # Handle auto_defer
            context._handle_auto_defer(self.auto_defer)

            self._discord.dispatch("slash_command", context)
            if hasattr(command, "invoke"):
                await command.invoke(context, **options)
            elif command.callback is not None:
                await command.callback(context, **options)
            return

    def _get_cog_commands(self, cog):
        # Get all BaseCallables flagged with __type__ of 1 (command)
        return [x[1] for x in inspect.getmembers(cog, lambda x: isinstance(x, BaseCallable) and x.__type__ == 1)]
    def add_build(self, builder):
        """Adds a subclass of `SlashBuilder` to the internal cache and creates the command in the api
        
        Parameters
        ----------
        builder: :class:`~SlashBuilder`
            The built SlashCommand you want to add
        
        """
        builder.register(self)

    def add_command(self, name=None, callback=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None) -> Union[SlashCommand, Coroutine]:
        """
        Adds a new slashcommand

        name: :class:`str`
            1-32 characters long name; default MISSING

            .. note::

                The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
        callback: :class:`function`, optional
            A callback that will be called when the command was received
        description: :class:`str`, optional
            1-100 character description of the command; default the command name
        options: List[:class:`~SlashOptions`], optional
            The parameters for the command; default MISSING
        choices: List[:class:`tuple`] | List[:class:`dict`], optional
            Choices for string and int types for the user to pick from; default MISSING
        guild_ids: List[:class:`str` | :class:`int`], optional
            A list of guild ids where the command is available; default MISSING
        default_permission: :class:`bool` | :class:`discord.Permissions`, optional
           Permissions that a user needs to have in order to execute the command, default ``True``.
                    If a bool was passed, it will indicate whether all users can use the command (``True``) or not (``False``)
        guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
            The permissions for the command in guilds
                Format: ``{"guild_id": SlashPermission}``

        Raises
        -------
        :class:`ClientException`
            Commands should be synced but the client is not ready yet
        """
        command = SlashCommand(callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, 
            guild_permissions=guild_permissions, state=self._discord._connection)
        return self.commands.add(command)
    def command(self, name=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None) -> Callable[..., SlashCommand]:
        """
        A decorator for a slash command
        
        command in discord:
            ``/name [options]``

        Parameters
        ----------
        name: :class:`str`, optional
            1-32 characters long name; default MISSING
            
            .. note::

                The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
        
        description: :class:`str`, optional
            1-100 character description of the command; default the command name
        options: List[:class:`~SlashOptions`], optional
            The parameters for the command; default MISSING
        choices: List[:class:`tuple`] | List[:class:`dict`], optional
            Choices for string and int types for the user to pick from; default MISSING
        guild_ids: List[:class:`str` | :class:`int`], optional
            A list of guild ids where the command is available; default MISSING
        default_permission: :class:`bool` | :class:`discord.Permissions`, optional
            Permissions that a user needs to have in order to execute the command, default ``True``.
                If a bool was passed, it will indicate whether all users can use the command (``True``) or not (``False``)
        guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
            The permissions for the command in guilds
                Format: ``{"guild_id": SlashPermission}``

        Decorator
        ---------
        callback: :class:`method(ctx)`
            The asynchron function that will be called if the command was used
                ctx: :class:`~SlashInteraction`
                    The used slash command

                .. note::

                    ``ctx`` is just an example name, you can use whatever you want for that

        Example
        -------
        .. code-block::

            @slash.command(name="hello_world", description="This is a test command", 
            options=[
                SlashOption(str, name="parameter", description="this is a parameter", choices=[("choice 1", "test")])
            ], guild_ids=[785567635802816595], default_permission=False, 
            guild_permissions={
                    785567635802816595: SlashPermission(allowed={"539459006847254542": SlashPermission.USER})
                }
            )
            async def command(ctx, parameter = None):
                ...
        """
        def wrapper(callback):
            return self.add_command(name, callback, description, options, guild_ids, default_permission, guild_permissions)
        return wrapper
    def add_subcommand(self, base_names, name=None, callback=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None):
        return self.commands.add(SlashSubcommand(callback, base_names, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions, state=self._discord._connection))
    def subcommand(self, base_names, name=None, description=None, options=None, guild_ids=None, default_permission=True, guild_permissions=None):
        """
        A decorator for a subcommand group
        
        command in discord
            ``/base_names... name [options]``

        Parameters
        ----------
        base_names: List[:class:`str`] | :class:`str`
            The names of the parent bases, currently limited to 2
                If you want to make a subcommand (``/base name``), you have to use a str instead of a list
        name: :class:`str`, optional
            1-32 characters long name; default MISSING
            
            .. note::

                The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
        description: :class:`str`, optional
            1-100 character description of the command; default the command name
        options: List[:class:`~SlashOptions`], optional
            The parameters for the command; default MISSING
        choices: List[:class:`tuple`] | List[:class:`dict`], optional
            Choices for string and int types for the user to pick from; default MISSING
        guild_ids: List[:class:`str` | :class:`int`], optional
            A list of guild ids where the command is available; default MISSING
        default_permission: :class:`bool` | :class:`discord.Permissions`, optional
            Permissions that a user needs to have in order to execute the command, default ``True``. 
                If a bool was passed, it will indicate whether all users can use the command (``True``) or not (``False``)
        guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
            The permissions for the command in guilds
                Format: ``{"guild_id": SlashPermission}``

        .. note::

            Permissions will be the same for every subcommand with the same base

        Decorator
        ---------
        callback: :class:`method(ctx)`
            The asynchron function that will be called if the command was used
                ctx: :class:`~SubSlashInteraction`
                    The used slash command

                .. note::

                    ``ctx`` is just an example name, you can use whatever you want for that
        
        Example
        -------
        
        subcommand

        .. code-block::

            @slash.subcommand_group(base_names="hello", name="world", options=[
                SlashOption(type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=[785567635802816595])
            async def command(ctx, user):
                ...

        subcommand-group

        .. code-block::

            @slash.subcommand_group(base_names=["hello", "beautiful"], name="world", options=[
                SlashOption(type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=[785567635802816595])
            async def command(ctx, user):
                ...

        """
        def wrapper(callback):
            return self.commands.add(SlashSubcommand(
                callback, base_names, name, description, options=options, 
                guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions,
                state=self._discord._connection
            ))
        return wrapper
    def user_command(self, name=None, guild_ids=None, default_permission=True, guild_permissions=None):
        """
        Decorator for user context commands in discord.
            ``Right-click username`` -> ``apps`` -> ``commands is displayed here``


        Parameters
        ----------
        name: :class:`str`, optional
            The name of the command; default MISSING
        guild_ids: List[:class:`str` | :class:`int`]
            A list of guilds where the command can be used
        default_permission: :class:`bool` | :class:`discord.Permissions`, optional
            Permissions that a user needs to have in order to execute the command, default ``True``.
                    If a bool was passed, it will indicate whether all users can use the command (``True``) or not (``False``)
        guild_permissions: Dict[:class:`SlashPermission`], optional
            Special permissions for guilds; default MISSING

        Decorator
        ---------
        callback: :class:`method(ctx, user)`
            The asynchron function that will be called if the command was used
                ctx: :class:`~SubSlashInteraction`
                    The used slash command
                user: :class:`discord.Member`
                    The user on which the command was used
                
                .. note::

                    ``ctx`` and ``user`` are just example names, you can use whatever you want for that

        Example
        -------
        
        .. code-block::

            @slash.user_command(name="call", guild_ids=[785567635802816595], default_permission=False, guild_permissions={
                785567635802816595: SlashPermission(allowed={
                    "585567635802816595": SlashPermission.USER
                })
            })
            async def call(ctx, message):
                ...
        """
        def wrapper(callback) -> UserCommand:
            return self.commands.add(UserCommand(callback, name, guild_ids, default_permission, guild_permissions, state=self._discord._connection))
        return wrapper
    def message_command(self, name=None, guild_ids=None, default_permission=True, guild_permissions=None):
        """
        Decorator for message context commands in discord.
            ``Right-click message`` -> ``apps`` -> ``commands is displayed here``


        Parameters
        ----------
        name: :class:`str`, optional
            The name of the command; default MISSING
        guild_ids: List[:class:`str` | :class:`int`]
            A list of guilds where the command can be used
        default_permission: :class:`bool` | :class:`discord.Permissions`, optional
            Permissions that a user needs to have in order to execute the command, default ``True``.
                If a bool was passed, it will indicate whether all users can use the command (``True``) or not (``False``)
        guild_permissions: Dict[:class:`SlashPermission`], optional
            Special permissions for guilds; default MISSING

        Decorator
        ---------
        callback: :class:`method(ctx, message)`
            The asynchron function that will be called if the command was used
                ctx: :class:`~SubSlashInteraction`
                    The used slash command
                message: :class:`~Message`
                    The message on which the command was used
            
                .. note::

                    ``ctx`` and ``message`` are just example names, you can use whatever you want for that
    
        Example
        -------
        
        .. code-block::

            @slash.message_command(name="quote", guild_ids=[785567635802816595], default_permission=False, guild_permissions={
                785567635802816595: SlashPermission(allowed={
                    "585567635802816595": SlashPermission.USER
                })
            })
            async def quote(ctx, message):
                ...
        """
        def wrapper(callback):
            return self.commands.add(MessageCommand(callback, name, guild_ids, default_permission, guild_permissions, state=self._discord._connection))
        return wrapper

class Components():
    """
    A class for using and receiving message components in discord
    
    Parameters
    -----------
    client: :class:`discord.Client`
        The main discord client

    override_dpy: :class:`bool`
        Whether some of discord.py's default methods should be overriden with this libary's; Default ``True``
            For more information see https://github.com/discord-py-ui/discord-ui/blob/main/discord_ui/override.py

    auto_defer: Tuple[:class:`bool`, :class:`bool`]
        Settings for the auto-defer; Default ``(True, False)``

        ``[0]``: Whether interactions should be deferred automatically

        ``[1]``: Whether the deferration should be hidden (True) or public (False)

    Example
    ------------------
    Example for using the listener

    
    .. code-block::

        ...
        # Your bot declaration should be here
        components = Components(client)
        
    
    for listening to button presses, use
    
    .. code-block::

        ...
        @client.event('on_button')
        async def on_button(ButtonInteraction):
            ...


    for listening to select menu selections, use

    .. code-block::

        ...
        @client.event('on_select')
        async def on_select(seletedMenu):
            ...

    For components that will listen to a custom id, use

    .. code-block::

        ...
        @components.listening_component(custom_id="custom_id_here")
        async def my_func(ctx):
            ...

    """
    def __init__(self, client: commands.Bot, override_dpy=True, auto_defer=False):
        """
        Creates a new compnent listener
        
        Example
        ```py
        Components(client, auto_defer=(True, False))
        ```
        """
        if override_dpy:
            override_it()

        self.auto_defer: Tuple[bool, bool] = (auto_defer, False) if isinstance(auto_defer, bool) else auto_defer
        self.listening_components: Dict[str, List[ListeningComponent]] = {}
        """A list of components that are listening for interaction"""
        self._discord: commands.Bot = client
        self._discord._connection._component_listeners = {}
        if discord.__version__.startswith("2"):
            self._discord.add_listener(self._on_component_response, "on_socket_raw_receive")
        elif discord.__version__.startswith("1"):
            self._discord.add_listener(self._on_component_response, 'on_socket_response')

        old_add = self._discord.add_cog
        def add_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            for com in self._get_listening_cogs(cog):
                com.cog = cog
                if self.listening_components.get(com.custom_id) is None:
                    self.listening_components[com.custom_id] = []
                self.listening_components[com.custom_id].append(com)
            old_add(*args, **kwargs)
        self._discord.add_cog = add_cog_override

        old_remove = self._discord.remove_cog
        def remove_cog_override(*args, **kwargs):
            cog = args[0] if len(args) > 0 else kwargs.get("cog")
            for com in self._get_listening_cogs(cog):
                com.cog = cog
                self.remove_listening_component(com)
            old_remove(*args, **kwargs)
        self._discord.remove_cog = remove_cog_override
    
    async def _on_component_response(self, msg):
        if discord.__version__.startswith("2"):
            if isinstance(msg, bytes):
                raise NotImplementedError("decompressing was removed! Please upgrade your discord.py version")
            if isinstance(msg, str):
                msg = json.loads(msg)
        
        if msg["t"] != "INTERACTION_CREATE":
            return
        data = msg["d"]
        
        if data["type"] != 3:
            return
        
        user = discord.Member(data=data["member"], guild=self._discord._connection._get_guild(int(data["guild_id"])), state=self._discord._connection) if data.get("member") is not None else discord.User(state=self._discord._connection, data=data["user"])
        if user.dm_channel is None:
            await user.create_dm()
        msg = await getMessage(self._discord._connection, data=data, response=True)
        
        interaction = Interaction(self._discord._connection, data, user, msg)
        if self.auto_defer[0] is True:
            await interaction.defer(self.auto_defer[1])
        self._discord.dispatch("interaction_received", interaction)

        self._discord.dispatch("component", ComponentContext(self._discord._connection, data, user, msg))


        # Handle auto_defer
        if int(data["data"]["component_type"]) == 2:
            for x in msg.components.buttons:
                if hasattr(x, 'custom_id') and x.custom_id == data["data"]["custom_id"]:
                    component = ButtonInteraction(data, user, x, msg, self._discord)
        elif int(data["data"]["component_type"]) == 3:
            for x in msg.components.selects:
                if x.custom_id == data["data"]["custom_id"]:
                    component = SelectInteraction(data, user, x, msg, self._discord)
        component._handle_auto_defer(self.auto_defer)
        
        
        # dispatch client events before listeners so the exception wont stop executing the function
        if ComponentType(data["data"]["component_type"]) is ComponentType.Button:
            self._discord.dispatch("button", component)
            self._discord.dispatch("button_press", component)   # deprecated
        elif ComponentType(data["data"]["component_type"]) is ComponentType.Select:
            self._discord.dispatch("select", component)
            self._discord.dispatch("menu_select", component)    # deprecated
        
        # Get listening components with the same custom id
        listening_components = self.listening_components.get(data["data"]["custom_id"])
        if listening_components is not None:
            for listening_component in listening_components:
                await listening_component.invoke(component)

        
        listener: Listener = self._discord._connection._component_listeners.get(str(msg.id))
        if listener is not None:
            await listener._call_listeners(component)


    async def send(self, channel, content=MISSING, *, tts=False, embed=MISSING, embeds=MISSING, file=MISSING, 
            files=MISSING, delete_after=MISSING, nonce=MISSING, allowed_mentions=MISSING, reference=MISSING, 
            mention_author=MISSING, components=MISSING) -> Message:
        """
        Sends a message to a textchannel

        Parameters
        ----------
        channel: :class:`discord.TextChannel` | :class:`int` | :class:`str`
            The target textchannel or the id of it
        content: :class:`str`, optional
            The message text content; default None
        tts: :class:`bool`, optional
            True if this is a text-to-speech message; default False
        embed: :class:`discord.Message`, optional
            Embedded rich content (up to 6000 characters)
        embeds: List[:class:`discord.Embed`], optional
            Up to 10 embeds; default None
        file: :class:`discord.File`, optional
            A file sent as an attachment to the message; default None
        files: List[:class:`discord.File`], optional
            A list of file attachments; default None
        delete_after: :class:`float`, optional
            After how many seconds the message should be deleted; default None
        nonce: :class:`int`, optional
            The nonce to use for sending this message. If the message was successfully sent, then the message will have a nonce with this value; default None
        allowed_mentions: :class:`discord.AllowedMentions`, optional
            A list of mentions proceeded in the message; default None
        reference: :class:`discord.MessageReference` | :class:`discord.Message`, optional
            A message to refer to (reply); default None
        mention_author: :class:`bool`, optional
            True if the author should be mentioned; default None
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`], optional
            A list of message components included in this message; default None

        Raises
        ------
        :class:`WrongType`
            Channel is not an instance of :class:`discord.abc.GuildChannel`, :class:`discord.abc.PrivateChannel:, :class:`int`, :class:`str` 


        Returns
        -------
        :class:`~Message`
            Returns the sent message
        """

        if not isinstance(channel, (discord.abc.GuildChannel, int, str, discord.User, discord.abc.PrivateChannel)):
            raise WrongType("channel", channel, ["discord.abc.PrivateChannel", "discord.abc.GuildChannel", "discord.User", "int"])

        channel_id = None
        if isinstance(channel, discord.User):
            if channel.dm_channel is None:
                channel = await channel.create_dm()
                channel_id = channel.id
            else:
                channel_id = channel.dm_channel
        elif isinstance(channel, discord.TextChannel):
            channel_id = channel.id
        else: 
            channel_id = channel
        payload = get_message_payload(content=content, tts=tts, embed=embed, embeds=embeds, nonce=nonce, allowed_mentions=allowed_mentions, reference=reference, mention_author=mention_author, components=components)

        route = BetterRoute("POST", f"/channels/{channel_id}/messages")

        r = None
        if file is MISSING and files is MISSING:
            r = await self._discord.http.request(route, json=payload)
        else:
            r = await send_files(route, files=_or(files, [file]), payload=payload, http=self._discord.http)

        msg = Message(state=self._discord._connection, channel=channel, data=r)
        
        if not _none(delete_after):
            await msg.delete(delay=delete_after)
        
        return msg
    def send_webhook(self, webhook, content=MISSING, *, wait=False, username=MISSING, avatar_url=MISSING, tts=False, files=MISSING, embed=MISSING, embeds=MISSING, allowed_mentions=MISSING, components=MISSING) -> Union[discord.WebhookMessage, None]:
        """
        Sends a webhook message
        
        Parameters
        ----------
        webhook: :class:`discord.Webhook`
            The webhook which will send the message
        content: :class:`str`, optional
            the message contents (up to 2000 characters); default None
        wait: :class:`bool`, optional
            if `True`, waits for server confirmation of message send before response, and returns the created message body; default False
        username: :class:`str`, optional
            override the default username of the webhook; default None
        avatar_url: :class:`str`, optional
            override the default avatar of the webhook; default None
        tts: :class:`bool`, optional
            true if this is a TTS message; default False
        files: :class:`discord.File`
            A list of files which will be sent as attachment
        embed: :class:`discord.Embed`
            Embed rich content, optional
        embeds: List[:class:`discord.Embed`], optional
            embedded rich content; default None
        allowed_mentions: :class:`discord.AllowedMentions`, optional
            allowed mentions for the message; default None
        components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`], optional
            the message components to include with the message; default None
        
        Returns
        -------
        :class:`~WebhookMessage` | :class:`None`
            The message which was sent, if wait was True, else nothing will be returned
        
        """
        payload = get_message_payload(content, tts=tts, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, components=components)
        payload["wait"] = wait
        if username is not None:
            payload["username"] = username
        if avatar_url is not None:
            payload["avatar_url"] = str(avatar_url)

        return webhook._adapter.execute_webhook(payload=payload, wait=wait, files=files)
    def listening_component(self, custom_id, messages=None, users=None, 
        component_type: Literal["button", "select"]=None,
        check: Callable[[Union[ButtonInteraction, SelectInteraction]], bool]=EMPTY_CHECK
    ):
        """
        Decorator for ``add_listening_component``

        Parameters
        ----------
        custom_id: :class:`str`
            The custom_id of the components to listen to
        messages: List[:class:`discord.Message` | :class:`int` :class:`str`], Optional
            A list of messages or message ids to filter the listening component
        users: List[:class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`], Optional
            A list of users or user ids to filter
        component_type: Literal[``'button'`` | ``'select'``]
            What type the used component has to be of (select: SelectMenu, button: Button)
        check: :class:`function`, Optional
            A function that has to return True in order to invoke the listening component
                The check function takes to parameters, the component and the message

        Decorator
        ---------
            callback: :class:`method(ctx)`
                The asynchron function that will be called if a component with the custom_id was invoked

                There will be one parameters passed

                    ctx: :class:`~ButtonInteraction` or :class:`~SelectInteraction`
                        The invoked component
                    
                    .. note::

                        ``ctx`` is just an example name, you can use whatever you want for it

        Example
        -------
        .. code-block::

            @ui.components.listening_component("custom_id", [539459006847254542], [53945900682362362])
            async def callback(ctx):
                ...
            
        """
        def wrapper(callback: Callable[[Union[ButtonInteraction, SelectInteraction]], Coroutine[Any, Any, Any]]):
            self.add_listening_component(callback, custom_id, messages, users, component_type, check)
        return wrapper
    def add_listening_component(self, callback, custom_id, messages=None, users=None, component_type: Literal["button", 2, "select", 3]=None, check: Callable[[Union[Component, Button, SelectMenu]], bool]=EMPTY_CHECK):
        """
        Adds a listener to received components

        Parameters
        ----------
        callback: :class:`function`
            The callback function that will be called when the component was received
        custom_id: :class:`str`
            The custom_id of the components to listen to
        messages: List[:class:`discord.Message` | :class:`int` :class:`str`], Optional
            A list of messages or message ids to filter the listening component
        users: List[:class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`], Optional
            A list of users or user ids to filter
        component_type: :class:`str` | :class:`int`
            The type of which the component has to be
        check: :class:`function`, Optional
            A function that has to return True in order to invoke the listening component
                The check function takes to parameters, the component and the message

        Raises
        -------
        :class:`MissingListenedComponentParameters`
            The callback for the listening component is missing required parameters
        :class:`NoAsyncCallback`
            The callback is not defined with the `async` keyword
        """
        if not inspect.iscoroutinefunction(callback):
            raise NoAsyncCallback()
        if len(inspect.signature(callback).parameters) < 1:
            raise MissingListenedComponentParameters()
        
        if self.listening_components.get(custom_id) is None:
            self.listening_components[custom_id] = []
        self.listening_components[custom_id].append(ListeningComponent(callback, messages, users, component_type, check, custom_id))
    def remove_listening_components(self, custom_id):
        """
        Removes all listening components for a custom_id
        
        Parameters
        ----------
        custom_id: :class:`str`
            The custom_id for the listening component
        
        """
        if self.listening_components.get(custom_id) is not None:
            del self.listening_components[custom_id]
    def remove_listening_component(self, listening_component):
        """
        Removes a listening component
        
        Parameters
        ----------
        listening_component: :class:`ListeningComponent`
            The listening component which should be removed
        
        """
        with contextlib.supress(KeyError): 
            self.listening_components[listening_component.custom_id].remove(listening_component)

    def _get_listening_cogs(self, cog):
        return [x[1] for x in inspect.getmembers(cog, lambda x: isinstance(x, ListeningComponent))]

    async def put_listener_to(self, target_message, listener):
        """Adds a listener to a message and edits it if the components are missing
        
        Parameters
        ----------
        target_message: :class:`Message`
            The message to which the listener should be attached
        listener: :class:`Listener`
            The listener which should be put to the message
        
        """
        if len(target_message.components) == 0:
            await target_message.edit(components=listener.to_components())
        self.attach_listener_to(target_message, listener)

    def attach_listener_to(self, target_message, listener):
        """Attaches a listener to a message after it was sent
        
        Parameters
        ----------
        target_message: :class:`Message`
            The message to which the listener should be attached
        listener: :class:`Listener`
            The listener that will be attached
        
        """
        listener._start(target_message, target_message)
    def clear_listeners(self):
        """Removes all component listeners"""
        self._connection._component_listeners = {}

class UI():
    """
    The main extension for the package to use slash commands and message components
        
        Parameters
        ----------
        client: :class:`discord.ext.commands.Bot`
            The discord bot client

        override_dpy: :class:`bool`
            Whether some of discord.py's default methods should be overriden with this libary's; Default ``True``
                For more information see https://github.com/discord-py-ui/discord-ui/blob/main/discord_ui/override.py

        slash_options: :class:`dict`, optional
            Settings for the slash command part; Default `{parse_method: ParseMethod.AUTO, delete_unused: False, wait_sync: 1}`
            
            ``parse_method``: :class:`int`, optional
                How the received interaction argument data should be treated; Default ``ParseMethod.AUTO``

            ``auto_sync``: :class:`bool`, optional
                Whether the libary should sync the slash commands automatically; Default ``True``

            ``delete_unused``: :class:`bool`, optional
                Whether the commands that are not registered by this slash ui should be deleted in the api; Default ``False``

            ``sync_on_cog``: :class:`bool`, optional
                Whether the slashcommands should be updated whenever a new cog is added or removed; Default ``True``

            ``wait_sync``: :class:`float`, optional
                How many seconds will be waited until the commands are going to be synchronized; Default ``1``

    auto_defer: Tuple[:class:`bool`, :class:`bool`]
        Settings for the auto-defer; Default ``(True, False)``

        ``[0]``: Whether interactions should be deferred automatically

        ``[1]``: Whether the deferration should be hidden (True) or public (False)
    """
    def __init__(self, client, override_dpy=True, slash_options = {"parse_method": ParseMethod.AUTO, "auto_sync": True, "delete_unused": False, "sync_on_cog": True, "wait_sync": 1}, auto_defer = False) -> None:
        """
        Creates a new ui object
        
        Example
        ```py
        UI(client, slash_options={"delete_unused": True, "wait_sync": 2}, auto_defer=True)
        ```
        """
        # enable debug events if needed
        if discord.__version__.startswith("2"):
            client._enable_debug_events = True

        self.components: Components = Components(client, override_dpy=override_dpy, auto_defer=auto_defer)
        """For using message components"""
        self.logger = logging
        if slash_options is None:
            slash_options = {"resolve_data": True, "delete_unused": False, "wait_sync": 1, "auto_defer": auto_defer}
        if slash_options.get("auto_defer") is None:
            slash_options["auto_defer"] = auto_defer
        self.slash: Slash = Slash(client, **slash_options)
        """For using slash commands"""
        
        # region shortcuts
        self.listening_component = self.components.listening_component
        self.command = self.slash.command
        self.subcommand = self.slash.subcommand
        self.message_command = self.slash.message_command
        self.user_command = self.slash.user_command
        # endregion



from discord.errors import ClientException

class NoCommandFound(ClientException):
    """Exception that is raised when you try to get a command with a name that doesn't exists"""
class AlreadyDeferred(ClientException):
    """Exception that is raised when you try to defer an interaction that was already deferred."""
    def __init__(self, *args: object) -> None:
        super().__init__("Interaction was already deferred")
class EphemeralDeletion(ClientException):
    """Exception that is raised when you try to delete an ephemeral message.
    
    Ephemeral messages can not be deleted"""
    def __init__(self, *args: object) -> None:
        super().__init__("Cannot delete an ephemeral message")
class MissingOptionParameter(ClientException):
    """Exception that is raised when a callback is missing a parameter which was 
    specified in the slash command.
    
    If you have a slashcommand with ``role`` as the name, your callback has to 
    accept a parameter with the name ``role``.

    For example
    
    .. code-block::

        @ui.slash.command(..., options=[SlashOption(SomeType, role, required=True)])
        async def callback(ctx, role):      # role is the name of the option
            ...
    """
    def __init__(self, option_name, *args: object) -> None:
        super().__init__("Missing parameter '" + option_name + "' in callback function")
class OptionalOptionParameter(ClientException):
    """Exception that is rarised when a callback function has a required parameter which 
    is marked optional in the slash command.
    
    If you want to have an optional option in your callback, you need to specify a default value 
    for it: ``async def callback(ctx, my_option=None)``
    """
    def __init__(self, param_name, *args: object) -> None:
        super().__init__("Parameter '" + param_name + "' in callback function needs to be optional (" + param_name + "=None)")
class NoAsyncCallback(ClientException):
    """Exception that is raised when a sync callback was provided
    
    Callbacks have to be async
    """
    def __init__(self, name) -> None:
        if name:
            msg = f"callback for command '{name}'' has to be async"
        else:
            msg = "callback has to be async"
        super().__init__(msg)
class CallbackMissingContextCommandParameters(ClientException):
    """Exception that is raised when a callback for a context command is missing parmeters.
    
    A context-command callback has to accept two parameters, one for the interaction context
    and the other one for the passed parameter.
    """
    def __init__(self, *args: object) -> None:
        super().__init__("Callback function for context commands has to accept 2 parameters (the used command, the message/user on which the interaction was used)")



"""
discord-ui extension
~~~~~~~~~~~~~~~~~~~~

A discord.py extension for discord's ui features like Buttons, SelectMenus, LinkButtons, 
slash-commands and context-commands (message-commands and user-commands).


- - -

Links
    [**Docs**](https://discord-ui.rtfd.io/) | [**Github**](https://github.com/discord-py-ui/discord-ui/) | [**PyPi**](https://pypi.org/project/discord-ui/) | [**License**](https://github.com/git/git-scm.com/blob/main/MIT-LICENSE.txt)

- Made by [404kuso](https://github.com/404kuso) and [RedstoneZockt](https://github.com/RedstoneZockt)
- Made for [discord.py](https://github.com/Rapptz/discord.py) and you

- - -

### Issues, Bugs, etc.

If you find any issues, bugs, problems or anything, please report them to our [github](https://github.com/discord-py-ui/discord-ui/issues/) so we can fix them

### Ideas

If you have ideas for this package, plz feel free to tell us

### Help

If you need any help or assist, join our [discord](https://discord.gg/bDJCGD994p)

"""


from .client import *
from .components import *
from .slash.types import *
from .slash.tools import *
from .tools import *
from .slash.tools import *
from .receive import *
from .listener import *
from .slash import ext
from .enums import ButtonStyle, OptionType, Channel, Mentionable


from .override import override_dpy


__title__ = "discord-ui"
__version__ = "5.1.0"
__author__ = "404kuso, RedstoneZockt"